/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1990 The Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#) portmap.c 1.6 96/07/06 23:06:23";
#endif /* not lint */

/*
@(#)portmap.c	2.3 88/08/11 4.0 RPCSRC
static char sccsid[] = "@(#)portmap.c 1.32 87/08/06 Copyr 1984 Sun Micro";
*/

/*
 * portmap.c, Implements the program,version to port number mapping for
 * rpc.
 */

/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdio.h>
#include <syslog.h>
#include <netdb.h>
#include <getopt.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>

#include <smallrpc/rpc.h>
#include <smallrpc/pmap_prot.h>

#include "pmap_check.h"
#include "rpcent.h"

static void	usage(int exval);
static int	makesock(int, int, unsigned long);
static void	reg_service(struct svc_req *, SVCXPRT *);
static void	drop_privilege(const char *, const char *);
static void	callit(struct svc_req *, SVCXPRT *);
static void	pml_create(unsigned int, unsigned int, int, unsigned int);
static void	pml_append(struct pmap *map);
static void	pml_load(void);
static void	pml_save(void);


static struct pmaplist *	pmaplist;
static int			debugging = 0;
static const char *		savefile = NULL;

 /*
  * How desperate can one be. It is possible to prevent an attacker from
  * manipulating your portmapper tables from outside with requests that
  * contain spoofed source address information. The countermeasure is to
  * force all rpc servers to register and unregister with the portmapper via
  * the loopback network interface, instead of via the primary network
  * interface that every host can talk to. For this countermeasure to work it
  * is necessary to #define LOOPBACK_SETUNSET, to disable source routing in
  * the kernel, and to modify libc so that get_myaddress() chooses the
  * loopback interface address.
  */
int
main(int argc, char **argv)
{
	char		*rootdir = NULL, *username = NULL;
	SVCXPRT		*xprt;
	int		sock, c;

	while ((c = getopt(argc, argv, "df:hr:u:v")) != EOF) {
		switch (c) {
		case 'd':
			debugging = 1;
			break;

		case 'f':
			savefile = optarg;
			break;

		case 'r':
			rootdir = optarg;
			break;

		case 'u':
			username = optarg;
			break;

		case 'v':
			verboselog = 1;
			break;

		case 'h':
			usage(0);

		default:
			usage(1);
		}
	}

	if (optind < argc)
		usage(1);

	if (rootdir && !username) {
		fprintf(stderr, "You cannot specify the chroot option -r without "
				"the username option -u.\n");
		exit(1);
	}

	openlog("portmap", debugging ? LOG_PID | LOG_PERROR : LOG_PID, LOG_DAEMON);

	/* Get list of rpc service names before we start up */
	rpcname_init();

	sock = makesock(SOCK_DGRAM, IPPROTO_UDP, INADDR_ANY);
	if ((xprt = svcudp_create(sock)) == NULL) {
		fprintf(stderr, "couldn't do udp_create");
		exit(1);
	}
	pml_create(PMAPPROG, PMAPVERS, IPPROTO_UDP, PMAPPORT);
	svc_register(xprt, PMAPPROG, PMAPVERS, reg_service, 0);

	sock = makesock(SOCK_STREAM, IPPROTO_TCP, INADDR_ANY);
	if ((xprt = svctcp_create(sock, RPCSMALLMSGSIZE, RPCSMALLMSGSIZE)) == NULL) {
		fprintf(stderr, "couldn't do tcp_create");
		exit(1);
	}
	pml_create(PMAPPROG, PMAPVERS, IPPROTO_TCP, PMAPPORT);
	svc_register(xprt, PMAPPROG, PMAPVERS, reg_service, 0);

#ifdef LOOPBACK_SETUNSET
	sock = makesock(SOCK_DGRAM, IPPROTO_UDP, INADDR_LOOPBACK);
	if ((xprt = svcudp_create(sock)) == NULL) {
		fprintf(stderr, "couldn't do udp_create");
		exit(1);
	}
	svc_register(xprt, PMAPPROG, PMAPVERS, reg_service, 0);

	sock = makesock(SOCK_STREAM, IPPROTO_TCP, INADDR_LOOPBACK);
	if ((xprt = svctcp_create(sock, RPCSMALLMSGSIZE, RPCSMALLMSGSIZE)) == NULL) {
		fprintf(stderr, "couldn't do tcp_create");
		exit(1);
	}
	svc_register(xprt, PMAPPROG, PMAPVERS, reg_service, 0);
#endif

	/* Drop all privilege and chroot to some safe directory */
	if (username)
		drop_privilege(rootdir, username);

	/* Go to the background */
	if (!debugging && daemon(0, 0)) {
		fprintf(stderr, "portmap: fork: %s", strerror(errno));
		exit(1);
	}

	/* Load old pmap list */
	if (savefile)
		pml_load();

	signal(SIGCHLD, SIG_IGN);

	svc_run();
	exit(1);
}

static void
usage(int exval)
{
	fprintf(stderr,
		"usage: portmap [-dv] [-u username] [-r rootdir] [-f filename]\n"
		"  -d        enable debugging\n"
		"  -v        verbose logging\n"
		"  -u user   run as the given user\n"
		"  -r dir    chroot to this directory\n"
		"  -f file   store port mapping in this file\n");
	exit(exval);
}

static int
makesock(int type, int proto, unsigned long addr)
{
	struct sockaddr_in sin;
	int		sock;

	if ((sock = socket(AF_INET, type, proto)) < 0) {
		fprintf(stderr, "cannot create udp socket: %m");
		exit(1);
	}

#ifdef LOOPBACK_SETUNSET
	{
	int	on = 1;
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on);
	}
#endif

	memset(&sin, 0, sizeof(sin));
	sin.sin_addr.s_addr = htonl(addr);
	sin.sin_family = AF_INET;
	sin.sin_port = htons(PMAPPORT);
	if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) != 0) {
		fprintf(stderr, "cannot bind %s: %m", (proto == IPPROTO_UDP)? "udp" : "tcp");
		exit(1);
	}

	return sock;
}

static void
drop_privilege(const char *rootdir, const char *username)
{
	struct passwd	*pwd;

	/* To be done */
	if (!(pwd = getpwnam(username))) {
		fprintf(stderr, "No such user: %s\n", username);
		exit(1);
	}

	if (rootdir
	 && (chdir(rootdir) < 0 || chroot(rootdir) < 0)) {
		fprintf(stderr, "failed to chroot to %s: %m", rootdir);
		exit(1);
	}

	if (setgroups(1, &pwd->pw_gid) < 0
	 || setgid(pwd->pw_gid) < 0
	 || setuid(pwd->pw_uid) < 0) {
		fprintf(stderr, "failed to drop privilege: %m\n");
		exit(1);
	}
}

static struct pmaplist *
pml_find(unsigned int prog, unsigned int vers, unsigned int prot)
{
	struct pmaplist *pml, *hit = NULL;

	for (pml = pmaplist; pml != NULL; pml = pml->pml_next) {
		if (pml->pml_map.pm_prog != prog
		 || pml->pml_map.pm_prot != prot)
			continue;
		hit = pml;
		if (pml->pml_map.pm_vers == vers)
		    break;
	}
	return (hit);
}

static void
pml_create(unsigned int prog, unsigned int vers, int prot, unsigned int port)
{
	struct pmap	map;

	map.pm_prog = prog;
	map.pm_vers = vers;
	map.pm_prot = prot;
	map.pm_port = port;
	pml_append(&map);
}

static void
pml_append(struct pmap *map)
{
	struct pmaplist	*pml, **pp;

	pml = (struct pmaplist *) malloc(sizeof(struct pmaplist));
	pml->pml_map  = *map;
	pml->pml_next = 0;

	for (pp = &pmaplist; *pp; pp = &(*pp)->pml_next)
		;
	*pp = pml;
}

static void
pml_load(void)
{
	FILE		*fp;
	struct pmap	map;
	int		n;

	if (!savefile || !(fp = fopen(savefile, "r")))
		return;

	while (1) {
		n = fscanf(fp, "%u %u %u %u",
				&map.pm_prog, &map.pm_vers,
				&map.pm_prot, &map.pm_port);
		if (n != 4)
			break;
		pml_append(&map);
	}
	fclose(fp);
}

static void
pml_save(void)
{
	struct pmaplist	*pml;
	FILE		*fp;

	if (!savefile)
		return;

	if ((fp = fopen(savefile, "w")) == NULL) {
		static int	complained = 0;

		if (!complained++)
			syslog(LOG_ERR, "unable to open %s: %m", savefile);
		return;
	}

	for (pml = pmaplist; pml; pml = pml->pml_next) {
		fprintf(fp,
			"%u %u %u %u\n",
			pml->pml_map.pm_prog,
			pml->pml_map.pm_vers,
			pml->pml_map.pm_prot,
			pml->pml_map.pm_port);
	}

	fclose(fp);
}

/*
 * Send response to PMAP call
 */
static inline int
doreply(SVCXPRT *xprt, xdrproc_t proc, void *ptr)
{
	if (!svc_sendreply(xprt, proc, ptr)) {
		if (debugging) {
			fprintf(stderr, "svc_sendreply failed\n");
			abort();
		}
		return 0;
	}
	return 1;
}

/* 
 * 1 OK, 0 not
 */
void
reg_service(struct svc_req *rqstp, SVCXPRT *xprt)
{
	struct sockaddr	*caller;
	struct pmap	reg;
	struct pmaplist *pml, *fnd;
	
	/*
	 * Later wrappers change the logging severity on the fly. Reset to
	 * defaults before handling the next request.
	 */
	allow_severity = LOG_INFO;
	deny_severity = LOG_WARNING;

	if (debugging)
		fprintf(stderr, "server: about to do a switch\n");

	caller = svc_getcaller(xprt);
	switch (rqstp->rq_proc) {
	case PMAPPROC_NULL:
		/*
		 * Null proc call
		 */
		/* remote host authorization check */
		check_default(rqstp, 0);
		doreply(xprt, (xdrproc_t) xdr_void, NULL);
		break;

	case PMAPPROC_SET:
		/*
		 * Set a program,version to port mapping
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_pmap, &reg))
			svcerr_decode(xprt);
		else {
			int	okay = 0;

			/* reject non-local requests, protect priv. ports */
			if (!check_setunset(rqstp, reg.pm_prog, reg.pm_port))
				goto done_set;

			/* Chuck out garbage requests */
			switch (reg.pm_prot) {
			case IPPROTO_UDP:
			case IPPROTO_TCP:
				if (reg.pm_port == 0
				 || reg.pm_port >= 65536)
					goto done_set;
				break;

			default:
				/* Reject non UDP TCP */
				goto done_set;
			}

			/*
			 * check to see if already used
			 * pml_find returns a hit even if
			 * the versions don't match, so check for it
			 */
			fnd = pml_find(reg.pm_prog, reg.pm_vers, reg.pm_prot);
			if (fnd && fnd->pml_map.pm_vers == reg.pm_vers) {
				if (fnd->pml_map.pm_port == reg.pm_port)
					okay = 1;
			} else {
				pml_append(&reg);
				pml_save();
				okay = 1;
			}

		done_set:
			doreply(xprt, (xdrproc_t) xdr_int, &okay);
		}
		break;

	case PMAPPROC_UNSET:
		/*
		 * Remove a program,version to port mapping.
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_pmap, &reg))
			svcerr_decode(xprt);
		else {
			struct pmaplist	**pp;
			int		okay = 0;

			/* reject non-local requests */
			if (!check_setunset(rqstp, reg.pm_prog, 0))
				goto done_unset;

			/* First check whether the caller's allowed to 
			 * remove all of the registrations, then do it. */
			for (pml = pmaplist; pml; pml = pml->pml_next) {
				if (pml->pml_map.pm_prog == reg.pm_prog
				 && pml->pml_map.pm_vers == reg.pm_vers
				 && !check_privileged_port(rqstp,
						pml->pml_map.pm_prog,
						pml->pml_map.pm_port))
					goto done_unset;
			}
			okay = 1;

			pp = &pmaplist;
			while ((pml = *pp) != NULL) {
				if (pml->pml_map.pm_prog == reg.pm_prog
				 && pml->pml_map.pm_vers == reg.pm_vers) {
					*pp = pml->pml_next;
					free(pml);
				} else {
					pp = &pml->pml_next;
				}
			}
			if (okay)
				pml_save();
		done_unset:
			doreply(xprt, (xdrproc_t) xdr_int, &okay);
		}
		break;

	case PMAPPROC_GETPORT:
		/*
		 * Lookup the mapping for a program,version and return its port
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_pmap, &reg))
			svcerr_decode(xprt);
		else {
			int	port = 0;

			/* remote host authorization check */
			if (check_default(rqstp, reg.pm_prog)
			 && (fnd = pml_find(reg.pm_prog, reg.pm_vers, reg.pm_prot)))
				port = fnd->pml_map.pm_port;
			doreply(xprt, (xdrproc_t) xdr_int, &port);
		}
		break;

	case PMAPPROC_DUMP:
		/*
		 * Return the current set of mapped program,version
		 */
		if (!svc_getargs(xprt, xdr_void, NULL))
			svcerr_decode(xprt);
		else {
			/* remote host authorization check */
			struct pmaplist *p = NULL;

			if (check_default(rqstp, 0))
				p = pmaplist;
			doreply(xprt, (xdrproc_t) xdr_pmaplist, &p);
		}
		break;

	case PMAPPROC_CALLIT:
		/*
		 * Calls a procedure on the local machine.  If the requested
		 * procedure is not registered this procedure does not return
		 * error information!!
		 * This procedure is only supported on rpc/udp and calls via 
		 * rpc/udp.  It passes null authentication parameters.
		 */
		callit(rqstp, xprt);
		break;

	default:
		/* remote host authorization check */
		check_default(rqstp, 0);
		svcerr_noproc(xprt);
		break;
	}
}


/*
 * Stuff for the rmtcall service
 */
#define ARGSIZE		(9000)

struct encap_parms {
	unsigned int		arglen;
	void *			args;
};

struct rmtcallargs {
	unsigned long		rmt_prog;
	unsigned long		rmt_vers;
	unsigned long		rmt_port;
	unsigned long		rmt_proc;
	struct encap_parms	rmt_args;
};

static bool_t
xdr_encap_parms(XDR *xdrs, struct encap_parms *epp)
{
	return xdr_bytes(xdrs, (void **) &epp->args, &epp->arglen, ARGSIZE);
}

static bool_t
xdr_rmtcall_args(XDR *xdrs, struct rmtcallargs *cap)
{
	return xdr_u_long(xdrs, &cap->rmt_prog) &&
	       xdr_u_long(xdrs, &cap->rmt_vers) &&
	       xdr_u_long(xdrs, &cap->rmt_proc) &&
	       xdr_encap_parms(xdrs, &cap->rmt_args);
}

static bool_t
xdr_rmtcall_result(XDR *xdrs, struct rmtcallargs *cap)
{
	return xdr_u_long(xdrs, &cap->rmt_port)
	    && xdr_encap_parms(xdrs, &cap->rmt_args);
	return 0;
}

/*
 * only worries about the struct encap_parms part of struct rmtcallargs.
 * The arglen must already be set!!
 */
static bool_t
xdr_opaque_parms(XDR *xdrs, struct rmtcallargs *cap)
{
	if (cap->rmt_args.arglen > ARGSIZE)
		return 0;
	return xdr_opaque(xdrs, cap->rmt_args.args, cap->rmt_args.arglen);
}

/*
 * This routine finds and sets the length of incoming opaque paraters
 * and then calls xdr_opaque_parms.
 */
static bool_t
xdr_len_opaque_parms(XDR *xdrs, struct rmtcallargs *cap)
{
	cap->rmt_args.arglen = xdr_getavail(xdrs);
	return xdr_opaque_parms(xdrs, cap);
}

/*
 * Call a remote procedure service
 * This procedure is very quiet when things go wrong.
 * The proc is written to support broadcast rpc.  In the broadcast case,
 * a machine should shut-up instead of complain, less the requestor be
 * overrun with complaints at the expense of not hearing a valid reply ...
 *
 * This now forks so that the program & process that it calls can call 
 * back to the portmapper.
 */
static void
callit(struct svc_req *rqstp, SVCXPRT *xprt)
{
	struct rmtcallargs a;
	struct pmaplist *pml;
	struct sockaddr_in me;
	int		xid, pid, so = -1, retry;
	struct rpc_msg	msg;
	char		argbuf[ARGSIZE],
			msgbuf[16000];
	XDR		xdrs;

	a.rmt_args.args = argbuf;
	if (!svc_getargs(xprt, (xdrproc_t) xdr_rmtcall_args, &a))
		return;

	/* host and service access control */
	if (!check_callit(rqstp, a.rmt_prog, a.rmt_proc))
		return;
	if ((pml = pml_find(a.rmt_prog, a.rmt_vers, IPPROTO_UDP)) == NULL)
		return;
	a.rmt_port = pml->pml_map.pm_port;

	/*
	 * fork a child to do the work.  Parent immediately returns.
	 * Child exits upon completion.
	 */
	if ((pid = fork()) != 0) {
		if (pid < 0)
			syslog(LOG_ERR, "CALLIT (prog %lu): fork: %m", a.rmt_prog);
		return;
	}

	if ((so = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
		exit(0);

	memset(&me, 0, sizeof(me));
	me.sin_family = AF_INET;
	me.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	me.sin_port = htons(pml->pml_map.pm_port);

	/* Manually assemble the RPC message, and copy the raw
	 * credentials supplied by the caller
	 */
	memset(&msg, 0, sizeof(msg));
	msg.rm_xid = xid = getpid();
	msg.rm_direction = CALL;
	msg.rm_call.cb_rpcvers = 2;
	msg.rm_call.cb_prog = a.rmt_prog;
	msg.rm_call.cb_vers = a.rmt_vers;
	msg.rm_call.cb_proc = a.rmt_proc;
	memcpy(&msg.rm_call.cb_cred, &rqstp->rq_cred, sizeof(struct opaque_auth));
	memcpy(&msg.rm_call.cb_verf, &rqstp->rq_verf, sizeof(struct opaque_auth));

	xdrmem_create(&xdrs, msgbuf, sizeof(msgbuf), XDR_ENCODE);
	if (!xdr_callmsg(&xdrs, &msg)
	 || !xdr_opaque(&xdrs, a.rmt_args.args, a.rmt_args.arglen))
		exit(0);

	/* Prepare parsing of the response */
	memset(&msg, 0, sizeof(msg));
	msg.rm_reply.rp_acpt.ar_results.proc = (xdrproc_t) xdr_len_opaque_parms;
	msg.rm_reply.rp_acpt.ar_results.where = argbuf;

	/* Now send the thing to the server and wait for a response */
	for (retry = 0; retry < 5; retry++) {
		struct timeval	timeout = { 5, 0 };
		fd_set		fset;
		int		n;

		if (sendto(so, msgbuf, xdr_getpos(&xdrs), 0,
				(struct sockaddr *) &me, sizeof(me)) < 0)
			exit(0);
		FD_ZERO(&fset);
		FD_SET(so, &fset);

		select(so+1, &fset, NULL, NULL, &timeout);
		if (FD_ISSET(so, &fset)
		 && (n = recv(so, msgbuf, sizeof(msgbuf), 0)) > 0) {
			xdrmem_create(&xdrs, msgbuf, n, XDR_DECODE);
			if (xdr_replymsg(&xdrs, &msg) && msg.rm_xid == xid) {
				/* XXX: copy the server's verifier into
				 * our response? */
				svc_sendreply(xprt,
					(xdrproc_t) xdr_rmtcall_result, &a);
				break;
			}
		}
	}

	exit(0);
}
